package com.tzld.piaoquan.recommend.server.service.filter.strategy;

import com.tzld.piaoquan.recommend.server.common.ThreadPoolFactory;
import com.tzld.piaoquan.recommend.server.service.filter.FilterParam;
import com.tzld.piaoquan.recommend.server.service.filter.FilterStrategy;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.dao.DataAccessException;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.redis.core.RedisOperations;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.SessionCallback;
import org.springframework.data.redis.core.SetOperations;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
@Component
@Slf4j
public class ViewedStrategy implements FilterStrategy {
    @Autowired
    @Qualifier("filterRedisTemplate")
    private RedisTemplate<String, String> redisTemplate;

    @Autowired
    private MongoTemplate mongoTemplate;

    private String keyFormat = "user:exclude:videoidset:%s";

    @Override
    public List<Long> filter(FilterParam param) {
        String user = StringUtils.isNotBlank(param.getUid()) ? param.getUid() : param.getMid();
        if (StringUtils.isBlank(user)) {
            return param.getVideoIds();
        }
        String key = String.format(keyFormat, user);
        Set<String> viewedVideoIds = redisTemplate.opsForSet().members(key);
        if (CollectionUtils.isEmpty(viewedVideoIds)) {
            // 从mongo取曝光数据
            Criteria criteria = new Criteria();
            criteria.and("uid").is(user);
            Query query = new Query();
            query.addCriteria(criteria);
            List<VideoView> list = mongoTemplate.find(query, VideoView.class);
            //TODO 为什么限制最多10000条？是不是限制近几天更合适？
            if (CollectionUtils.isNotEmpty(list)) {
                int limit = 10000;
                for (int i = list.size() - 1; i >= 0 && limit-- > 0; i--) {
                    viewedVideoIds.add(String.valueOf(list.get(i).getVideoId()));
                }

                // 异步写Redis
                ThreadPoolFactory.defaultPool().execute(() -> {
                    redisTemplate.executePipelined(new SessionCallback<String>() {
                        @Override
                        public <A, B> String execute(RedisOperations<A, B> redisOperations) throws DataAccessException {
                            SetOperations<String, String> operations =
                                    (SetOperations<String, String>) redisOperations.opsForSet();
                            operations.add(key, viewedVideoIds.toArray(new String[viewedVideoIds.size()]));
                            redisTemplate.expire(key, 360 * 3600, TimeUnit.SECONDS);
                            return null;
                        }
                    });
                });
            }
        }

        if (CollectionUtils.isEmpty(viewedVideoIds)) {
            return param.getVideoIds();
        }
        return param.getVideoIds().stream()
                .filter(vid -> !viewedVideoIds.contains(String.valueOf(vid)))
                .collect(Collectors.toList());
    }

}
